/* $Id: aprspacket.cpp 361 2010-01-17 10:38:10Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 *
 *
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Own stuff.
#include "aprspacket.h"

// Wacky stuff.
#include <fap.h>

// std stuff.
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <bitset>
using std::bitset;
#include <sstream>
using std::ostringstream;
using std::istringstream;
#include <cstring>



// STATIC MEMBER FUNCTIONS


AprsPacket* AprsPacket::fromTnc2Packet(string const& packet)
{
	// Remove badness.
	string input;
	for ( unsigned int i = 0; i < packet.length(); ++i )
	{
		if ( packet.at(i) == '\r' || packet.at(i) == '\n' )
		{
			break;
		}
		input += packet.at(i);
	}
	
	// Try to parse.
	fap_packet_t* data = fap_parseaprs(input.c_str(), input.length(), 1);
	if ( data->error_code )
	{
		fap_free(data);
		data = fap_parseaprs(input.c_str(), input.length(), 0);
		return new AprsPacket(data, false);
	}
	
	return new AprsPacket(data, true);
}


AprsPacket* AprsPacket::fromAx25Frame(string const& ax25frame)
{
	// Conversion constants.
	int const FEND = 0xc0;
	int const FESC = 0xdb;
	int const TFEND = 0xdc;
	int const TFESC = 0xdd;

	// Perform byte stuffing into KISS frame.
	string kissframe;
	for ( unsigned int i = 0; i < ax25frame.size(); ++i )
	{
		if ( ax25frame.at(i) == FEND || ax25frame.at(i) == FESC )
		{
			kissframe.append(1, FESC);
			if ( ax25frame.at(i) == FEND )
			{
				kissframe.append(1, TFEND);
			}
			else
			{
				kissframe.append(1, TFESC);
			}
		}
		else
		{
			kissframe += ax25frame.at(i);
		}
	}
	
	// Use KISS parser to do the trick.
	return fromKissFrame(ax25frame);
}


AprsPacket* AprsPacket::fromKissFrame(string const& kissframe)
{
	// Convert to TNC-2 using libfap.
	unsigned int packet_len = 512, tnc_id = 666;
	char buffer[packet_len];
	memset(buffer, 0, packet_len);
	if ( !fap_kiss_to_tnc2(kissframe.c_str(), kissframe.length(), buffer, &packet_len, &tnc_id) )
	{
		return new AprsPacket(kissframe, buffer);
	}

	// Remove badness.
	string packet;
	for ( unsigned int i = 0; i < packet_len; ++i )
	{
		if ( buffer[i] == '\r' || buffer[i] == '\n' )
		{
			break;
		}
		packet += buffer[i];
	}
	
	// Return result.
	fap_packet_t* data = fap_parseaprs(packet.c_str(), packet.length(), 1);
	return new AprsPacket(data, true);
}



// CONSTRUCTORS


// From given details.
AprsPacket::AprsPacket(std::string const& source, std::string const& target, vector<std::string> const& path, std::string const& body, bool const& is_ax25):
errorMessage_(), origPacket_(), data_(NULL), is_ax25_(is_ax25), timestamp_(time(NULL))
{
	// Construct the packet in TNC-2 format.
	ostringstream outflow;
	outflow << source << ">" << target;
	for ( unsigned int i = 0; i < path.size(); i++ )
	{
		outflow << "," << path.at(i);
	}   
	outflow << ":" << body;
	
	// Remove badness.
	string tmp_packet = outflow.str();
	for ( unsigned int i = 0; i < tmp_packet.length(); ++i )
	{
		if ( tmp_packet.at(i) == '\r' || tmp_packet.at(i) == '\n' )
		{
			break;
		}
		origPacket_ += tmp_packet.at(i);
	}
	
	// Parse the packet.
	data_ = fap_parseaprs(origPacket_.c_str(), origPacket_.size(), is_ax25);
	
	// Check for errors.
	if ( data_->error_code )
	{
		errorMessage_ = fap_explain_error(*data_->error_code);
	}
}


// Copy constructor.
AprsPacket::AprsPacket(AprsPacket const& orig, vector<string> const& extraPath):
errorMessage_(orig.errorMessage_), origPacket_(orig.origPacket_), data_(NULL), is_ax25_(orig.is_ax25_), timestamp_(orig.timestamp_)
{
	// Don't adjust anything for invalid packets.
	if ( errorMessage_.empty() )
	{
		// Check if we need to append stuff to path.
		if ( !extraPath.empty() )
		{
			// Create copy of orig's original packet.
			string newpkt (orig.origPacket_);
			
			// Check if it has path which can be lenghtened.
			string::size_type pos = newpkt.find(":");
			if ( pos != string::npos && pos > 0 )
			{
				// Create extrapath string.
				string extrapath;
				for (unsigned int i = 0; i < extraPath.size(); ++i )
				{
					extrapath += "," + extraPath.at(i);
				}
				// Push extrapath into end of original path.
				newpkt.insert(pos, extrapath);
			}
			
			// Save the actual packet.
			origPacket_ = newpkt;
		}
		
		// Create copy of data.
		data_ = fap_parseaprs(origPacket_.c_str(), origPacket_.size(), is_ax25_);
		
		// If we adjusted path, it might went wrong.
		if ( !extraPath.empty() )
		{
			// If we were AX.25 compatible, try to get valid again by degenetaring to non-AX.25-compatible.
			if ( is_ax25_ && data_->error_code )
			{
				fap_free(data_);
				is_ax25_ = 0;
				data_ = fap_parseaprs(origPacket_.c_str(), origPacket_.size(), is_ax25_);
			}
			
			// If we're hopelessly invalid, set error message.
			if ( data_->error_code )
			{
				errorMessage_ = fap_explain_error(*data_->error_code);
			}
		}
	}
}



AprsPacket::~AprsPacket()
{
	fap_free(data_);
}
	

// GETTERS

string AprsPacket::getTnc2Packet() const
{
	// Convert to printable and return it if we're not valid.
	if ( !errorMessage_.empty() )
	{
		string result;
		char buf[5];
		for ( unsigned int i = 0;i < origPacket_.size(); ++i )
		{
			if ( !isprint(origPacket_.at(i)) )
			{
				memset(buf, 0, 5);
				sprintf(buf, "<%02X>", origPacket_.at(i) & 0xFF);
				result += buf;
			}
			else
			{
				result += origPacket_.at(i);
			}
		}
		return result + " (" + errorMessage_ + ") ";
	}
	
	// By default, we return it as-is.
	return origPacket_;
}
	


string AprsPacket::getKissFrame() const
{
	// In case of error we return empty.
	if ( !errorMessage_.empty() )
	{
		return string();
	}
	
	// Convert using libfap.
	unsigned int output_len = 512;
	char buffer[output_len];
	memset(buffer, 0, output_len);
	if ( fap_tnc2_to_kiss(origPacket_.c_str(), origPacket_.length(), 0, buffer, &output_len) )
	{
		return string(buffer, output_len);
	}

	return string();
}



string AprsPacket::getAsThirdParty(string const& owncall, string const& path) const
{
	if ( !errorMessage_.empty() )
	{
		return string();
	}
	
	string result = "}";
	result += data_->src_callsign;
	result += ">";
	result += data_->dst_callsign;
	result += "," + path + "*," + owncall + "*" + ":";
	result += string(data_->body, data_->body_len);
	
	return result;
}



AprsPacket* AprsPacket::getCarriedThirdParty() const
{
	if ( !errorMessage_.empty() )
	{
		return NULL;
	}
	
	if ( data_->body_len > 1 && data_->body[0] == '}' )
	{
		return fromTnc2Packet(string(data_->body+1, data_->body_len-1));
	}
	
	return NULL;
}



time_t AprsPacket::getTimestamp() const
{
	return timestamp_;
}



string AprsPacket::getSource() const
{
	if ( !errorMessage_.empty() )
	{
		return string();
	}

	if ( data_->src_callsign )
	{
		return string(data_->src_callsign);
	}
	
	return string();
}



string AprsPacket::getLabel() const
{
	if ( !errorMessage_.empty() )
	{
		return string();
	}
	
	if ( data_->type && (*data_->type == fapOBJECT || *data_->type == fapITEM) &&
		  data_->object_or_item_name )
	{
		return string(data_->object_or_item_name);
	}

	if ( data_->src_callsign )
	{
		return string(data_->src_callsign);
	}
	
	return string();
}



bool AprsPacket::getLocation(Location& loc) const
{
	if ( !errorMessage_.empty() )
	{
		return false;
	}
	
	if ( data_->latitude != NULL && data_->longitude != NULL )
	{
		loc.lon = *data_->longitude;
		loc.lat = *data_->latitude;
		return true;
	}
	return false;
}



bool AprsPacket::getMessage(string& message, string& destination) const
{
	if ( !errorMessage_.empty() )
	{
		return false;
	}
	
	if ( data_->type && *data_->type == fapMESSAGE &&
		  data_->message && data_->destination )
	{
		message = data_->message;
		destination = data_->destination;
		return true;
	}
	
	return false;
}



//OPERATORS

bool AprsPacket::error(string& message) const
{
	message = errorMessage_;
	return !message.empty();
}
	


bool AprsPacket::canBeGated() const
{
	if ( !errorMessage_.empty() )
	{
		return false;
	}
	
	// Check path for elements which disable gating of this packet.
	vector<string> path (data_->path, data_->path + data_->path_len);
	for ( unsigned int i = 0; i < path.size(); i++ )
	{
		if ( path.at(i).find("NOGATE") != string::npos )
		{
			return false;
		}
	}
	
	// Check for thidparty packets and queries.
	if ( data_->body_len > 0 && (data_->body[0] == '}' || data_->body[0] == '?') )
	{
		return false;
	}
	
	// No gatestoppers found.
	return true;
}



bool AprsPacket::canBeGatedToAprsIs() const
{
	if ( !errorMessage_.empty() )
	{
		return false;
	}
	
	// Check path for elements which disable gating of this packet.
	vector<string> path (data_->path, data_->path + data_->path_len);
	for ( unsigned int i = 0; i < path.size(); i++ )
	{
		if ( path.at(i).find("NOGATE") != string::npos || path.at(i).find("RFONLY") != string::npos ||
			  path.at(i).find("TCPIP") != string::npos || path.at(i).find("TCPXX") != string::npos )
		{
			return false;
		}
	}
	
	// Check for thidparty packets and queries.
	if ( data_->body_len > 0 && (data_->body[0] == '}' || data_->body[0] == '?') )
	{
		return false;
	}
	
	// No gatestoppers found.
	return true;
}


	
bool AprsPacket::operator==(const AprsPacket& theOther) const
{
	if ( !errorMessage_.empty() )
	{
		return false;
	}
	
	string::size_type i;
	
	string owntgt (data_->dst_callsign);
	if ( (i = owntgt.find("-")) != string::npos )
	{
		owntgt = owntgt.substr(0, i);
	}
	string othtgt (theOther.data_->dst_callsign);
	if ( (i = othtgt.find("-")) != string::npos )
	{
		othtgt = othtgt.substr(0, i);
	}
	
	return ( string(data_->src_callsign) == string(theOther.data_->src_callsign) &&
				owntgt == othtgt &&
				string(data_->body, data_->body_len) == string(theOther.data_->body, theOther.data_->body_len)
			 );
}

  
	
// PRIVATE


AprsPacket::AprsPacket(fap_packet_t* data, bool const& is_ax25):
errorMessage_(), origPacket_(data->orig_packet, data->orig_packet_len), data_(data), is_ax25_(is_ax25), timestamp_(time(0))
{
	if ( data_->error_code )
	{
		errorMessage_ = fap_explain_error(*data->error_code);
	}
}



AprsPacket::AprsPacket(string const& origPacket, string const& errorMessage):
errorMessage_(errorMessage), origPacket_(origPacket), data_(NULL), is_ax25_(false), timestamp_(0)
{
}
