/* $Id: beacon.cpp 363 2010-01-17 16:19:11Z oh2gve $
 * 
 * Copyright 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// own stuff.
#include "beacon.h"
#include "aprspacket.h"
#include "connection.h"

// std stuff.
#include <vector>
using std::vector;
#include <string>
using std::string;
#include <map>
using std::multimap;
using std::make_pair;
#include <ctime>
#include <cmath>
#include <sstream>
using std::ostringstream;

// wx stuff.
#include <wx/string.h>



Beacon::Beacon(std::vector<Connection*> const& conns, Connection* uplink, Cfg::Configuration const& cfg, DupeFilter& dupefilter, Packetlog& packetlog):
beacons_(), dupefilter_(dupefilter), packetlog_(packetlog),
outcons_(conns.size()+1, '_'), error_()
{
	// Beacon packet body is same for all connections, so we generate it first.

	// Check if we do tx or only rx.
	char overlay = 'R';
	if ( uplink->inputEnabled() )
	{
		overlay = 'I';
	}

	// Separate latitude into whole degrees and minutes.
	int lat_deg = fabs(cfg.location.lat);
	double lat_min = (fabs(cfg.location.lat) - lat_deg) * 60.0;

	// Separate longitude into whole degrees and minutes.
	int lon_deg = fabs(cfg.location.lon);
	double lon_min = (fabs(cfg.location.lon) - lon_deg) * 60.0;
	
	// Construct the packet.
	ostringstream info;
	info << "!";
	info << string(wxString::Format(wxT("%02d%02.2f"), lat_deg, lat_min).ToAscii());
	if ( cfg.location.lat < 0 )
	{
		info << "S";
	}
	else
	{
		info << "N";
	}
	info << overlay;
	info << string(wxString::Format(wxT("%03d%02.2f"), lon_deg, lon_min).ToAscii());
	if ( cfg.location.lon < 0 )
	{
		info << "W";
	}
	else
	{
		info << "E";
	}
	info << "&";
	info << cfg.comment.ToUTF8();

	// Create uplink beacon packet if necessary.
	if ( uplink->outputEnabled() )
	{
		// Create path.
		vector<string> netpath;
		netpath.push_back("TCPIP*");
		netpath.push_back("qAI");
		string tmp;
		tmp = cfg.callsign.ToAscii();
		netpath.push_back(tmp);
		
		// Create the packet.
		BeaconPacket bp;
		bp.packet = new AprsPacket(tmp, OWN_TARGET, netpath, info.str(), false);
		bp.conn = uplink;
		bp.connNo = 0;
		bp.interval = cfg.beaconInterval;
		
		// Check that we got valid packet.
		if ( bp.packet->error(tmp) )
		{
			error_ = string("Failed to create beacon packet for uplink: ") + tmp + "(" + bp.packet->getTnc2Packet() + ")";
		}
		
		// Schedule beaconing to happen ASAP.
		beacons_.insert( make_pair(0, bp) );
		
		
	}
	
	// Create RF beacons as necessary.
	if ( error_.empty() )
	{
		for ( unsigned int c = 0; c < cfg.connections.size(); ++c )
		{
			// Check if this conn does TX.
			if ( cfg.connections.at(c).state != Cfg::READONLY )
			{
				// Create path.
				vector<string> rfpath;
				for ( unsigned int i = 0; i < cfg.connections.at(c).path.size(); i++ )
				{
					rfpath.push_back(string(cfg.connections.at(c).path.at(i).ToAscii()));
				}
				
				// Create the packet.
				BeaconPacket bp;
				bp.packet = new AprsPacket(string(cfg.callsign.ToAscii()), OWN_TARGET, rfpath, info.str(), true);
				bp.conn = conns.at(c);
				bp.connNo = c+1;
				bp.interval = cfg.beaconInterval;
				
				// Check that we got valid packet.
				string tmp;
				if ( bp.packet->error(tmp) )
				{
					ostringstream errmsg;
					errmsg << "Failed to create beacon packet for connection " << c+1 << ": " << tmp;
					error_ = errmsg.str();
					break;
				}
				
				// Schedule beaconing to happen ASAP.
				beacons_.insert( make_pair(0, bp) );
			}
		}
	}
}



Beacon::~Beacon()
{
	for ( multimap<time_t, BeaconPacket>::iterator i = beacons_.begin();
			i != beacons_.end(); ++i )
	{
		delete(i->second.packet);
	}
}



bool Beacon::operator!() const
{
	return !error_.empty();
}


string Beacon::error() const
{
	return error_;
}



void Beacon::kick()
{
	// Check if next beaconing is at hand.
	if ( !beacons_.empty() && time(NULL) >= beacons_.begin()->first )
	{
		// Put the packet in dupe control.
		dupefilter_.isDupe(beacons_.begin()->second.packet);
		
		// Create a copy of the packet to be transmitted.
		AprsPacket* output = new AprsPacket(*beacons_.begin()->second.packet);
		
		// Transmit the packet.
		beacons_.begin()->second.conn->write(output);
		
		// Create a copy of the packet for packet log.
		Packet* logPkt = new Packet();
		logPkt->input = '_';
		logPkt->aprsPacket = new AprsPacket(*beacons_.begin()->second.packet);
		logPkt->thirdParty = NULL;
		logPkt->outputs = outcons_;
		if ( beacons_.begin()->second.connNo == 0 )
		{
			logPkt->outputs[beacons_.begin()->second.connNo] = 'u';
		}
		else
		{
			logPkt->outputs[beacons_.begin()->second.connNo] = static_cast<char>(beacons_.begin()->second.connNo+48);
		}
		logPkt->flag = 'b';
		
		// Put in log.
		packetlog_.write(logPkt);
		
		// Re-schedule the beacon.
		beacons_.insert( make_pair(time(NULL)+beacons_.begin()->second.interval, beacons_.begin()->second) );
		
		// Unschedule this beacon.
		beacons_.erase(beacons_.begin());
	}
}
