/* $Id: guard.cpp 358 2010-01-16 06:51:37Z oh2gve $
 * 
 * Copyright 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// own stuff.
#include "guard.h"
#include "packet.h"
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <list>
using std::list;
#include <map>
using std::map;
using std::make_pair;
#include <vector>
using std::vector;
#include <ctime>

// wx stuff.
#include <wx/log.h>



Guard::Guard(Connection* myconn, Cfg::ConnectionCfg const& mycfg, unsigned int const& mynum, unsigned int const& concount, DupeFilter& dupefilter,Packetlog& packetlog, string const& owncall):
myconn_(myconn), mycfg_(mycfg), mynum_(mynum+48), outcons_(concount, '_'),
dupefilter_(dupefilter), packetlog_(packetlog), owncall_(owncall), hold_(), clients_(),
packetCount_(0), timer_(time(NULL))
{
}



Guard::~Guard()
{
	while ( !hold_.empty() )
	{
		delete hold_.front()->aprsPacket;
		delete hold_.front();
		hold_.pop_front();
	}
}



Packet* Guard::givePacket(unsigned int const& timeout)
{
	if ( myconn_->inputEnabled() )
	{
		AprsPacket* aprspacket = myconn_->readTimeout(timeout);
		if ( aprspacket )
		{
			// Create result object.
			Packet* result = new Packet;
			result->input = mynum_;
			result->aprsPacket = aprspacket;
			result->thirdParty = NULL;
			result->outputs = outcons_;
			result->flag = '-';
				
			// Check if the packet is valid.
			string msg;
			if ( aprspacket->error(msg) )
			{
				// Print the packet.
				result->flag = 'e';
				packetlog_.write(result);
				return NULL;
			}
			
			// Try to get third-party packet from the packet.
			result->thirdParty = aprspacket->getCarriedThirdParty();
			while ( result->thirdParty )
			{
				// Update pointer to innermost packet.
				aprspacket = result->thirdParty;
				
				// Get next packet.
				AprsPacket* next3p = result->thirdParty->getCarriedThirdParty();
				
				// If we got someting, we move to it.
				if ( next3p )
				{
					delete result->thirdParty;
					result->thirdParty = next3p;
					aprspacket = next3p;
				}
				else
				{
					break;
				}
			}
			
			// Drop possible copy of this packet from our output hold.
			for ( list<Packet const*>::iterator i = hold_.begin(); i != hold_.end(); ++i )
			{
				if ( *aprspacket == *(*i)->aprsPacket )
				{
					delete (*i)->aprsPacket;
					delete *i;
					hold_.erase(i);
					break;
				}
			}
			
			// Check if the packet seen on RF passes our filter. Objects, items or stations seen only in third party format are not our clients.
			if ( !result->thirdParty && aprspacket->getSource() == aprspacket->getLabel() && passesFilter(aprspacket) )
			{
				// Save the target into our client list.
				clients_[aprspacket->getLabel()] = time(NULL);
			}
			else if ( !result->thirdParty )
			{
				// Does not pass, try to remove the target from client list. It may be there, if it has moved recently or something.
				for ( map<string, time_t>::iterator i = clients_.begin(); i != clients_.end(); ++i )
				{
					if ( i->first == aprspacket->getLabel() )
					{
						clients_.erase(i);
						break;
					}
				}
			}
			
			// Check for duplicateness.
			if ( dupefilter_.isDupe(aprspacket) )
			{
				result->flag = 'd';
				packetlog_.write(result);
				return NULL;
			}
			
			// Check if the packet can be gated.
			if ( !aprspacket->canBeGated() )
			{
				result->flag = 'n';
				packetlog_.write(result);
				return NULL;
			}
			
			// Return the packet.
			return result;
		}
	}
	
	return NULL;
}

bool Guard::passesFilter(AprsPacket const* packet) const
{
	// Check if packet is a message packet.
	string message, destination;
	if ( packet->getMessage(message, destination) )
	{
		// Check if destination is our client.
		for ( map<string, time_t>::const_iterator i = clients_.begin(); i != clients_.end(); ++i )
		{
			if ( i->first == destination )
			{
				return true;
			}
		}
	}
	
	// Check if packet contains location.
	AprsPacket::Location loc;
	bool hasLocation = false;
	if ( packet->getLocation(loc) )
	{
		hasLocation = true;
	}
	
	// Check packet agains filter elements.
	for ( unsigned int i = 0; i < mycfg_.filterElements.size(); ++i )
	{
		if ( hasLocation )
		{
			if ( mycfg_.filterElements.at(i).accepts(packet->getLabel(), loc.lon, loc.lat) )
			{
				return true;
			}
		}
		else
		{
			if ( mycfg_.filterElements.at(i).accepts(packet->getLabel()) )
			{
				return true;
			}
		}      
	}

	// Every element said no-go.
	return false;
}


void Guard::sendPacket(Packet* packet)
{
	hold_.push_back(packet);
}



void Guard::kick()
{
	if ( !hold_.empty() )
	{
		// Check timestamp of the oldest packet.
		if ( time(0) - hold_.front()->aprsPacket->getTimestamp() > mycfg_.limitercfg.holdtime )
		{
			// Okay, it's now old enough to be transmitted.
			
			// First we turn the packet into third party format.
			string packet ( hold_.front()->aprsPacket->getAsThirdParty( owncall_, hold_.front()->inputName ) );
			
			// Next we get path we're going to use for the carrying frame.
			vector<string> outpath;
			for ( unsigned int i = 0; i < mycfg_.path.size(); ++i )
			{
				outpath.push_back( string(mycfg_.path.at(i).ToAscii()) );
			}
			
			// Create ourselves a packet with the outgoing packet as body.
			AprsPacket* aprspacket = new AprsPacket( owncall_, string(OWN_TARGET), outpath, packet, true );
			string msg;
			if ( aprspacket->error(msg) )
			{
				wxLogMessage(wxT(" debug: connection") + wxString::Format(wxT("%c"), mynum_) + wxT(" failed to construct outgoing packet (") + wxString::FromAscii(msg.c_str()) + wxT(")"));
				delete aprspacket;
				return;
			}
			
			// Now we create a log copy of the packet.
			Packet* logPkt = new Packet(*hold_.front());
			logPkt->aprsPacket = new AprsPacket(*aprspacket);
			logPkt->outputs = outcons_;

			// Check if we're below maxrate.
			if ( packetCount_ < mycfg_.limitercfg.amountOfPackets )
			{
				// Yes we are, let's send the packet.
				myconn_->write(aprspacket);
				
				// Increase count.
				packetCount_++;
				
				// Mark the packet sent.
				logPkt->outputs.at(static_cast<unsigned int>(mynum_)-48) = mynum_;
			}
			else
			{
				// We're above maxrate, let's just delete the packet.
				delete aprspacket;
				
				// Mark the packet limited.
				logPkt->outputs.at(static_cast<unsigned int>(mynum_)-48) = 'l';
			}
			
			// Save packet into log with info of it's fate.
			packetlog_.write(logPkt);
			
			// Packet processed.
			delete hold_.front()->aprsPacket;
			delete hold_.front();
			hold_.pop_front();
		}
	}
	
	// Check if limit time cycle has completed.
	if ( myconn_->outputEnabled() && time(NULL) - timer_ >= mycfg_.limitercfg.amountOfTime )
	{
		// Cycle is complete, reset packet counter and start new cycle.
		packetCount_ = 0;
		timer_ = time(NULL);
	}
	
	// Flush client list.
	for ( map<string, time_t>::iterator i = clients_.begin(); i != clients_.end(); ++i )
	{
		// Check if last info from the client is older than 30 mins.
		if ( time(NULL) - i->second > 60*30 )
		{
			clients_.erase(i);
			// This function is run often enough to remove just one client at a time.
			break;
		}
	}
}
