/* $Id: main.h 358 2010-01-16 06:51:37Z oh2gve $
 *
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

/**
 * \mainpage
 * \section modules Modules
 * Main module is one big chunck containg almost all things. In addition
 * there's two much smaller modules: Config is a configuration file parser
 * and storage. Packet contains a model of an APRS packet and some
 * validators for it.
 * \section threads Threads
 * aprsg has a main thread, and one thread for each connection. Main
 * thread handles all of the logic and controls the other threads.
 */

/**
 * \todo Add "const" to many places.
 * \todo Check headers, specially copyright years.
 */

#ifndef MAIN_H
#define MAIN_H


// Own stuff.
#include "connection.h"
#include "uplink.h"
#include "cfg.h"

// std stuff.
#include <fstream>
#include <vector>
#include <string>

// wx stuff.
#include <wx/app.h>
#include <wx/file.h>


/// Main class of the application.
class AprsApp : public wxAppConsole
{
public:
	
	/// Constructor.
	/**
	 * Just initialized private members into sensible values.
	 */
	AprsApp();
	
	/// Command line parameter initialisation.
	/**
	 * Our constant table wxCmdLineEntryDesc CMDLINE[] is binded to the
	 * parser handled by wxWidgets.
	 */
	void OnInitCmdLine(wxCmdLineParser& parser);

	/// Save results of command line parser.
	/**
	 * Command line parser handles help requests itself without calling this
	 * function, which handles params daemon mode and configuration file
	 * location. Configuration file parsing is also done here.
	 */
	bool OnCmdLineParsed(wxCmdLineParser& parser);

	/// Initialization for main.
	/**
	 * Binds the signal handling function handleSignal() to SIGINT and SIGTERM.
	 */
	bool OnInit();

	/// The "int main()" function.
	/**
	 * After all initialization procedures are complete, this function is
	 * called by the wgWidgets to start the main operation. First thing
	 * done here is forking into a daemon process, if requested.
	 */
	int OnRun();

	/// Cleanup.
	/**
	 * wxWidgets calls this after OnRun(), but before it's own cleanup.
	 */
	int OnExit();

private:

	// Configuration data

	/// Cmd line flag telling wether to fork to backround or stay foreground.
	bool daemonMode_;

	/// Configuration file name.
	/**
	 * Read from command line. If not given, defaults to build-time SYSCONFDIR/aprsg.conf.
	 */
	wxString configFileName_;
	
	/// Contents of the config file.
	Cfg::Configuration configuration_;


	// Support

	/// Calculates aprspass of given callsign-ssid.
	std::string aprsPass(wxString callsign);

	/// Create us a pid file.
	bool createPidFile();
	
	/// Remove our pid file.
	void removePidFile();
	
	/// Pid file handle.
	wxFile* pidFile_;

	/// Open the general log file.
	bool openLog(bool const& initialOpening);
	
	/// Close general log.
	void closeLog();
	
	/// Reopen general log.
	bool reopenLog();
	
	/// Handle of general log file.
	std::ofstream* logfilePtr_;
	
	/// openLog() disconnects stdio by redirecting them here.
	int nullFd_;

	/// Open connections.
	/**
	 * Leaves previous successfully opened connections open when opening of
	 * another one fails.
	 * \return True if every connection was successfully opened. False if a
	 * connection failed to open.
	 */
	bool openConnections();
	
	/// Close all connections.
	void closeConnections();
	
	/// Server connection.
	Uplink* uplink_;
	
	/// Connection storage.
	std::vector<Connection*> connections_;

};

#endif // MAIN_H
